@@ -1,5 +1,5 @@
 # sqlalchemy/ext/baked.py
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -13,18 +13,20 @@ compiled result to be fully cached.
 
 """
 
-from ..orm.query import Query
-from ..orm import strategies, attributes, properties, \
-    strategy_options, util as orm_util, interfaces
-from .. import log as sqla_log
-from ..sql import util as sql_util, func, literal_column
-from ..orm import exc as orm_exc
-from .. import exc as sa_exc
-from .. import util
-
 import copy
 import logging
 
+from .. import exc as sa_exc
+from .. import util
+from ..orm import exc as orm_exc
+from ..orm import strategy_options
+from ..orm.query import Query
+from ..orm.session import Session
+from ..sql import func
+from ..sql import literal_column
+from ..sql import util as sql_util
+
+
 log = logging.getLogger(__name__)
 
 
@@ -39,7 +41,8 @@ class Bakery(object):
 
 
     """
-    __slots__ = 'cls', 'cache'
+
+    __slots__ = "cls", "cache"
 
     def __init__(self, cls_, cache):
         self.cls = cls_
@@ -52,7 +55,7 @@ class Bakery(object):
 class BakedQuery(object):
     """A builder object for :class:`.query.Query` objects."""
 
-    __slots__ = 'steps', '_bakery', '_cache_key', '_spoiled'
+    __slots__ = "steps", "_bakery", "_cache_key", "_spoiled"
 
     def __init__(self, bakery, initial_fn, args=()):
         self._cache_key = ()
@@ -142,21 +145,33 @@ class BakedQuery(object):
          :class:`.BakedQuery` object subsequent to the spoil step will be
          non-cached; the state of the :class:`.BakedQuery` up until
          this point will be pulled from the cache.   If True, then the
-         entire :class:`.Query` object is built from scratch each
+         entire :class:`_query.Query` object is built from scratch each
          time, with all creational functions being called on each
          invocation.
 
         """
         if not full and not self._spoiled:
             _spoil_point = self._clone()
-            _spoil_point._cache_key += ('_query_only', )
+            _spoil_point._cache_key += ("_query_only",)
             self.steps = [_spoil_point._retrieve_baked_query]
         self._spoiled = True
         return self
 
-    def _with_lazyload_options(self, options, effective_path, cache_path=None):
-        """Cloning version of _add_lazyload_options.
+    def _effective_key(self, session):
+        """Return the key that actually goes into the cache dictionary for
+        this :class:`.BakedQuery`, taking into account the given
+        :class:`.Session`.
+
+        This basically means we also will include the session's query_class,
+        as the actual :class:`_query.Query` object is part of what's cached
+        and needs to match the type of :class:`_query.Query` that a later
+        session will want to use.
+
         """
+        return self._cache_key + (session._query_cls,)
+
+    def _with_lazyload_options(self, options, effective_path, cache_path=None):
+        """Cloning version of _add_lazyload_options."""
         q = self._clone()
         q._add_lazyload_options(options, effective_path, cache_path=cache_path)
         return q
@@ -179,32 +194,37 @@ class BakedQuery(object):
         if cache_path.path[0].is_aliased_class:
             # paths that are against an AliasedClass are unsafe to cache
             # with since the AliasedClass is an ad-hoc object.
-            self.spoil()
+            self.spoil(full=True)
         else:
             for opt in options:
                 cache_key = opt._generate_cache_key(cache_path)
                 if cache_key is False:
-                    self.spoil()
+                    self.spoil(full=True)
                 elif cache_key is not None:
                     key += cache_key
 
         self.add_criteria(
-            lambda q: q._with_current_path(effective_path).
-            _conditional_options(*options),
-            cache_path.path, key
+            lambda q: q._with_current_path(
+                effective_path
+            )._conditional_options(*options),
+            cache_path.path,
+            key,
         )
 
     def _retrieve_baked_query(self, session):
-        query = self._bakery.get(self._cache_key, None)
+        query = self._bakery.get(self._effective_key(session), None)
         if query is None:
             query = self._as_query(session)
-            self._bakery[self._cache_key] = query.with_session(None)
+            self._bakery[self._effective_key(session)] = query.with_session(
+                None
+            )
         return query.with_session(session)
 
     def _bake(self, session):
         query = self._as_query(session)
 
         context = query._compile_context()
+
         self._bake_subquery_loaders(session, context)
         context.session = None
         context.query = query = context.query.with_session(None)
@@ -215,12 +235,74 @@ class BakedQuery(object):
         # so delete some compilation-use-only attributes that can take up
         # space
         for attr in (
-                '_correlate', '_from_obj', '_mapper_adapter_map',
-                '_joinpath', '_joinpoint'):
+            "_correlate",
+            "_from_obj",
+            "_mapper_adapter_map",
+            "_joinpath",
+            "_joinpoint",
+        ):
             query.__dict__.pop(attr, None)
-        self._bakery[self._cache_key] = context
+
+        # if the query is not safe to cache, we still do everything as though
+        # we did cache it, since the receiver of _bake() assumes subqueryload
+        # context was set up, etc.
+        if context.query._bake_ok:
+            self._bakery[self._effective_key(session)] = context
+
         return context
 
+    def to_query(self, query_or_session):
+        """Return the :class:`_query.Query` object for use as a subquery.
+
+        This method should be used within the lambda callable being used
+        to generate a step of an enclosing :class:`.BakedQuery`.   The
+        parameter should normally be the :class:`_query.Query` object that
+        is passed to the lambda::
+
+            sub_bq = self.bakery(lambda s: s.query(User.name))
+            sub_bq += lambda q: q.filter(
+                User.id == Address.user_id).correlate(Address)
+
+            main_bq = self.bakery(lambda s: s.query(Address))
+            main_bq += lambda q: q.filter(
+                sub_bq.to_query(q).exists())
+
+        In the case where the subquery is used in the first callable against
+        a :class:`.Session`, the :class:`.Session` is also accepted::
+
+            sub_bq = self.bakery(lambda s: s.query(User.name))
+            sub_bq += lambda q: q.filter(
+                User.id == Address.user_id).correlate(Address)
+
+            main_bq = self.bakery(
+                lambda s: s.query(Address.id, sub_bq.to_query(q).as_scalar())
+            )
+
+        :param query_or_session: a :class:`_query.Query` object or a class
+         :class:`.Session` object, that is assumed to be within the context
+         of an enclosing :class:`.BakedQuery` callable.
+
+
+         .. versionadded:: 1.3
+
+
+        """
+
+        if isinstance(query_or_session, Session):
+            session = query_or_session
+        elif isinstance(query_or_session, Query):
+            session = query_or_session.session
+            if session is None:
+                raise sa_exc.ArgumentError(
+                    "Given Query needs to be associated with a Session"
+                )
+        else:
+            raise TypeError(
+                "Query or Session object expected, got %r."
+                % type(query_or_session)
+            )
+        return self._as_query(session)
+
     def _as_query(self, session):
         query = self.steps[0](session)
 
@@ -237,10 +319,10 @@ class BakedQuery(object):
         a "baked" query so that we save on performance too.
 
         """
-        context.attributes['baked_queries'] = baked_queries = []
+        context.attributes["baked_queries"] = baked_queries = []
         for k, v in list(context.attributes.items()):
             if isinstance(v, Query):
-                if 'subquery' in k:
+                if "subquery" in k:
                     bk = BakedQuery(self._bakery, lambda *args: v)
                     bk._cache_key = self._cache_key + k
                     bk._bake(session)
@@ -248,15 +330,20 @@ class BakedQuery(object):
                 del context.attributes[k]
 
     def _unbake_subquery_loaders(
-            self, session, context, params, post_criteria):
+        self, session, context, params, post_criteria
+    ):
         """Retrieve subquery eager loaders stored by _bake_subquery_loaders
         and turn them back into Result objects that will iterate just
         like a Query object.
 
         """
+        if "baked_queries" not in context.attributes:
+            return
+
         for k, cache_key, query in context.attributes["baked_queries"]:
-            bk = BakedQuery(self._bakery,
-                            lambda sess, q=query: q.with_session(sess))
+            bk = BakedQuery(
+                self._bakery, lambda sess, q=query: q.with_session(sess)
+            )
             bk._cache_key = cache_key
             q = bk.for_session(session)
             for fn in post_criteria:
@@ -272,7 +359,8 @@ class Result(object):
     against a target :class:`.Session`, and is then invoked for results.
 
     """
-    __slots__ = 'bq', 'session', '_params', '_post_criteria'
+
+    __slots__ = "bq", "session", "_params", "_post_criteria"
 
     def __init__(self, bq, session):
         self.bq = bq
@@ -288,7 +376,8 @@ class Result(object):
         elif len(args) > 0:
             raise sa_exc.ArgumentError(
                 "params() takes zero or one positional argument, "
-                "which is a dictionary.")
+                "which is a dictionary."
+            )
         self._params.update(kw)
         return self
 
@@ -301,13 +390,14 @@ class Result(object):
         """Add a criteria function that will be applied post-cache.
 
         This adds a function that will be run against the
-        :class:`.Query` object after it is retrieved from the
+        :class:`_query.Query` object after it is retrieved from the
         cache.    Functions here can be used to alter the query in ways
         that **do not affect the SQL output**, such as execution options
         and shard identifiers (when using a shard-enabled query object)
 
         .. warning::  :meth:`.Result.with_post_criteria` functions are applied
-           to the :class:`.Query` object **after** the query's SQL statement
+           to the :class:`_query.Query`
+           object **after** the query's SQL statement
            object has been retrieved from the cache.   Any operations here
            which intend to modify the SQL should ensure that
            :meth:`.BakedQuery.spoil` was called first.
@@ -332,7 +422,7 @@ class Result(object):
         if not self.session.enable_baked_queries or bq._spoiled:
             return iter(self._as_query())
 
-        baked_context = bq._bakery.get(bq._cache_key, None)
+        baked_context = bq._bakery.get(bq._effective_key(self.session), None)
         if baked_context is None:
             baked_context = bq._bake(self.session)
 
@@ -341,7 +431,8 @@ class Result(object):
         context.attributes = context.attributes.copy()
 
         bq._unbake_subquery_loaders(
-            self.session, context, self._params, self._post_criteria)
+            self.session, context, self._params, self._post_criteria
+        )
 
         context.statement.use_labels = True
         if context.autoflush and not context.populate_existing:
@@ -355,7 +446,7 @@ class Result(object):
     def count(self):
         """return the 'count'.
 
-        Equivalent to :meth:`.Query.count`.
+        Equivalent to :meth:`_query.Query.count`.
 
         Note this uses a subquery to ensure an accurate count regardless
         of the structure of the original statement.
@@ -364,7 +455,7 @@ class Result(object):
 
         """
 
-        col = func.count(literal_column('*'))
+        col = func.count(literal_column("*"))
         bq = self.bq.with_criteria(lambda q: q.from_self(col))
         return bq.for_session(self.session).params(self._params).scalar()
 
@@ -373,7 +464,7 @@ class Result(object):
         if no rows present.  If multiple rows are returned,
         raises MultipleResultsFound.
 
-        Equivalent to :meth:`.Query.scalar`.
+        Equivalent to :meth:`_query.Query.scalar`.
 
         .. versionadded:: 1.1.6
 
@@ -389,13 +480,15 @@ class Result(object):
     def first(self):
         """Return the first row.
 
-        Equivalent to :meth:`.Query.first`.
+        Equivalent to :meth:`_query.Query.first`.
 
         """
         bq = self.bq.with_criteria(lambda q: q.slice(0, 1))
         ret = list(
-            bq.for_session(self.session).params(self._params).
-            _using_post_criteria(self._post_criteria))
+            bq.for_session(self.session)
+            .params(self._params)
+            ._using_post_criteria(self._post_criteria)
+        )
         if len(ret) > 0:
             return ret[0]
         else:
@@ -404,14 +497,18 @@ class Result(object):
     def one(self):
         """Return exactly one result or raise an exception.
 
-        Equivalent to :meth:`.Query.one`.
+        Equivalent to :meth:`_query.Query.one`.
 
         """
         try:
             ret = self.one_or_none()
-        except orm_exc.MultipleResultsFound:
-            raise orm_exc.MultipleResultsFound(
-                "Multiple rows were found for one()")
+        except orm_exc.MultipleResultsFound as err:
+            util.raise_(
+                orm_exc.MultipleResultsFound(
+                    "Multiple rows were found for one()"
+                ),
+                replace_context=err,
+            )
         else:
             if ret is None:
                 raise orm_exc.NoResultFound("No row was found for one()")
@@ -421,7 +518,7 @@ class Result(object):
         """Return one or zero results, or raise an exception for multiple
         rows.
 
-        Equivalent to :meth:`.Query.one_or_none`.
+        Equivalent to :meth:`_query.Query.one_or_none`.
 
         .. versionadded:: 1.0.9
 
@@ -435,12 +532,13 @@ class Result(object):
             return None
         else:
             raise orm_exc.MultipleResultsFound(
-                "Multiple rows were found for one_or_none()")
+                "Multiple rows were found for one_or_none()"
+            )
 
     def all(self):
         """Return all rows.
 
-        Equivalent to :meth:`.Query.all`.
+        Equivalent to :meth:`_query.Query.all`.
 
         """
         return list(self)
@@ -448,7 +546,7 @@ class Result(object):
     def get(self, ident):
         """Retrieve an object based on identity.
 
-        Equivalent to :meth:`.Query.get`.
+        Equivalent to :meth:`_query.Query.get`.
 
         """
 
@@ -471,13 +569,18 @@ class Result(object):
             # None present in ident - turn those comparisons
             # into "IS NULL"
             if None in primary_key_identity:
-                nones = set([
-                    _get_params[col].key for col, value in
-                    zip(mapper.primary_key, primary_key_identity)
-                    if value is None
-                ])
+                nones = set(
+                    [
+                        _get_params[col].key
+                        for col, value in zip(
+                            mapper.primary_key, primary_key_identity
+                        )
+                        if value is None
+                    ]
+                )
                 _lcl_get_clause = sql_util.adapt_criterion_to_null(
-                    _lcl_get_clause, nones)
+                    _lcl_get_clause, nones
+                )
 
             _lcl_get_clause = q._adapt_clause(_lcl_get_clause, True, False)
             q._criterion = _lcl_get_clause
@@ -494,16 +597,20 @@ class Result(object):
         # key so that if a race causes multiple calls to _get_clause,
         # we've cached on ours
         bq = bq._clone()
-        bq._cache_key += (_get_clause, )
+        bq._cache_key += (_get_clause,)
 
         bq = bq.with_criteria(
-            setup, tuple(elem is None for elem in primary_key_identity))
+            setup, tuple(elem is None for elem in primary_key_identity)
+        )
 
-        params = dict([
-            (_get_params[primary_key].key, id_val)
-            for id_val, primary_key
-            in zip(primary_key_identity, mapper.primary_key)
-        ])
+        params = dict(
+            [
+                (_get_params[primary_key].key, id_val)
+                for id_val, primary_key in zip(
+                    primary_key_identity, mapper.primary_key
+                )
+            ]
+        )
 
         result = list(bq.for_session(self.session).params(**params))
         l = len(result)
@@ -516,7 +623,8 @@ class Result(object):
 
 
 @util.deprecated(
-    "1.2", "Baked lazy loading is now the default implementation.")
+    "1.2", "Baked lazy loading is now the default implementation."
+)
 def bake_lazy_loaders():
     """Enable the use of baked queries for all lazyloaders systemwide.
 
@@ -528,18 +636,20 @@ def bake_lazy_loaders():
 
 
 @util.deprecated(
-    "1.2", "Baked lazy loading is now the default implementation.")
+    "1.2", "Baked lazy loading is now the default implementation."
+)
 def unbake_lazy_loaders():
     """Disable the use of baked queries for all lazyloaders systemwide.
 
-    This method now raises NotImplmentedError() as the "baked" implementation
+    This method now raises NotImplementedError() as the "baked" implementation
     is the only lazy load implementation.  The
-    :paramref:`.relationship.bake_queries` flag may be used to disable
+    :paramref:`_orm.relationship.bake_queries` flag may be used to disable
     the caching of queries on a per-relationship basis.
 
     """
     raise NotImplementedError(
-        "Baked lazy loading is now the default implementation")
+        "Baked lazy loading is now the default implementation"
+    )
 
 
 @strategy_options.loader_option()
@@ -553,20 +663,27 @@ def baked_lazyload(loadopt, attr):
 
 @baked_lazyload._add_unbound_fn
 @util.deprecated(
-    "1.2", "Baked lazy loading is now the default "
-    "implementation for lazy loading.")
+    "1.2",
+    "Baked lazy loading is now the default "
+    "implementation for lazy loading.",
+)
 def baked_lazyload(*keys):
     return strategy_options._UnboundLoad._from_keys(
-        strategy_options._UnboundLoad.baked_lazyload, keys, False, {})
+        strategy_options._UnboundLoad.baked_lazyload, keys, False, {}
+    )
 
 
 @baked_lazyload._add_unbound_all_fn
 @util.deprecated(
-    "1.2", "Baked lazy loading is now the default "
-    "implementation for lazy loading.")
+    "1.2",
+    "Baked lazy loading is now the default "
+    "implementation for lazy loading.",
+)
 def baked_lazyload_all(*keys):
     return strategy_options._UnboundLoad._from_keys(
-        strategy_options._UnboundLoad.baked_lazyload, keys, True, {})
+        strategy_options._UnboundLoad.baked_lazyload, keys, True, {}
+    )
+
 
 baked_lazyload = baked_lazyload._unbound_fn
 baked_lazyload_all = baked_lazyload_all._unbound_all_fn
