@@ -1,5 +1,5 @@
 # util/langhelpers.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -16,6 +16,7 @@ import itertools
 import operator
 import re
 import sys
+import textwrap
 import types
 import warnings
 
@@ -64,7 +65,10 @@ class safe_reraise(object):
             exc_type, exc_value, exc_tb = self._exc_info
             self._exc_info = None  # remove potential circular references
             if not self.warn_only:
-                compat.reraise(exc_type, exc_value, exc_tb)
+                compat.raise_(
+                    exc_value,
+                    with_traceback=exc_tb,
+                )
         else:
             if not compat.py3k and self._exc_info and self._exc_info[1]:
                 # emulate Py3K's behavior of telling us when an exception
@@ -75,7 +79,13 @@ class safe_reraise(object):
                     "is:\n %s %s\n" % (self._exc_info[0], self._exc_info[1])
                 )
             self._exc_info = None  # remove potential circular references
-            compat.reraise(type_, value, traceback)
+            compat.raise_(value, with_traceback=traceback)
+
+
+def clsname_as_plain_name(cls):
+    return " ".join(
+        n.lower() for n in re.findall(r"([A-Z][a-z]+)", cls.__name__)
+    )
 
 
 def decode_slice(slc):
@@ -121,8 +131,9 @@ def decorator(target):
     """A signature-matching decorator factory."""
 
     def decorate(fn):
-        if not inspect.isfunction(fn):
+        if not inspect.isfunction(fn) and not inspect.ismethod(fn):
             raise Exception("not a decoratable function")
+
         spec = compat.inspect_getfullargspec(fn)
         names = tuple(spec[0]) + spec[1:3] + (fn.__name__,)
         targ_name, fn_name = _unique_symbols(names, "target", "fn")
@@ -152,27 +163,31 @@ def _exec_code_in_env(code, env, fn_name):
     return env[fn_name]
 
 
-def public_factory(target, location):
+def public_factory(target, location, class_location=None):
     """Produce a wrapping function for the given cls or classmethod.
 
     Rationale here is so that the __init__ method of the
     class can serve as documentation for the function.
 
     """
+
     if isinstance(target, type):
         fn = target.__init__
         callable_ = target
         doc = (
-            "Construct a new :class:`.%s` object. \n\n"
+            "Construct a new :class:`%s` object. \n\n"
             "This constructor is mirrored as a public API function; "
-            "see :func:`~%s` "
+            "see :func:`sqlalchemy%s` "
             "for a full usage and argument description."
-            % (target.__name__, location)
+            % (
+                class_location if class_location else ".%s" % target.__name__,
+                location,
+            )
         )
     else:
         fn = callable_ = target
         doc = (
-            "This function is mirrored; see :func:`~%s` "
+            "This function is mirrored; see :func:`sqlalchemy%s` "
             "for a description of arguments." % location
         )
 
@@ -191,12 +206,43 @@ def %(name)s(%(args)s):
     env = {"cls": callable_, "symbol": symbol}
     exec(code, env)
     decorated = env[location_name]
-    decorated.__doc__ = fn.__doc__
+
+    if hasattr(fn, "_linked_to"):
+        linked_to, linked_to_location = fn._linked_to
+        linked_to_doc = linked_to.__doc__
+        if class_location is None:
+            class_location = "%s.%s" % (target.__module__, target.__name__)
+
+        linked_to_doc = inject_docstring_text(
+            linked_to_doc,
+            ".. container:: inherited_member\n\n    "
+            "This documentation is inherited from :func:`sqlalchemy%s`; "
+            "this constructor, :func:`sqlalchemy%s`,   "
+            "creates a :class:`sqlalchemy%s` object.  See that class for "
+            "additional details describing this subclass."
+            % (linked_to_location, location, class_location),
+            1,
+        )
+        decorated.__doc__ = linked_to_doc
+    else:
+        decorated.__doc__ = fn.__doc__
+
     decorated.__module__ = "sqlalchemy" + location.rsplit(".", 1)[0]
+    if decorated.__module__ not in sys.modules:
+        raise ImportError(
+            "public_factory location %s is not in sys.modules"
+            % (decorated.__module__,)
+        )
+
     if compat.py2k or hasattr(fn, "__func__"):
         fn.__func__.__doc__ = doc
+        if not hasattr(fn.__func__, "_linked_to"):
+            fn.__func__._linked_to = (decorated, location)
     else:
         fn.__doc__ = doc
+        if not hasattr(fn, "_linked_to"):
+            fn._linked_to = (decorated, location)
+
     return decorated
 
 
@@ -242,6 +288,26 @@ class PluginLoader(object):
         self.impls[name] = load
 
 
+def _inspect_func_args(fn):
+    try:
+        co_varkeywords = inspect.CO_VARKEYWORDS
+    except AttributeError:
+        # https://docs.python.org/3/library/inspect.html
+        # The flags are specific to CPython, and may not be defined in other
+        # Python implementations. Furthermore, the flags are an implementation
+        # detail, and can be removed or deprecated in future Python releases.
+        spec = compat.inspect_getfullargspec(fn)
+        return spec[0], bool(spec[2])
+    else:
+        # use fn.__code__ plus flags to reduce method call overhead
+        co = fn.__code__
+        nargs = co.co_argcount
+        return (
+            list(co.co_varnames[:nargs]),
+            bool(co.co_flags & co_varkeywords),
+        )
+
+
 def get_cls_kwargs(cls, _set=None):
     r"""Return the full set of inherited kwargs for the given `cls`.
 
@@ -250,7 +316,10 @@ def get_cls_kwargs(cls, _set=None):
     to pass along unrecognized keywords to its base classes, and the
     collection process is repeated recursively on each of the bases.
 
-    Uses a subset of inspect.getargspec() to cut down on method overhead.
+    Uses a subset of inspect.getfullargspec() to cut down on method overhead,
+    as this is used within the Core typing system to create copies of type
+    objects which is a performance-sensitive operation.
+
     No anonymous tuple arguments please !
 
     """
@@ -267,7 +336,7 @@ def get_cls_kwargs(cls, _set=None):
     )
 
     if has_init:
-        names, has_kw = inspect_func_args(ctr)
+        names, has_kw = _inspect_func_args(ctr)
         _set.update(names)
 
         if not has_kw and not toplevel:
@@ -282,26 +351,6 @@ def get_cls_kwargs(cls, _set=None):
     return _set
 
 
-try:
-    # TODO: who doesn't have this constant?
-    from inspect import CO_VARKEYWORDS
-
-    def inspect_func_args(fn):
-        co = fn.__code__
-        nargs = co.co_argcount
-        names = co.co_varnames
-        args = list(names[:nargs])
-        has_kw = bool(co.co_flags & CO_VARKEYWORDS)
-        return args, has_kw
-
-
-except ImportError:
-
-    def inspect_func_args(fn):
-        names, _, has_kw, _ = compat.inspect_getargspec(fn)
-        return names, bool(has_kw)
-
-
 def get_func_kwargs(func):
     """Return the set of legal kwargs for the given `func`.
 
@@ -310,7 +359,7 @@ def get_func_kwargs(func):
 
     """
 
-    return compat.inspect_getargspec(func)[0]
+    return compat.inspect_getfullargspec(func)[0]
 
 
 def get_callable_argspec(fn, no_self=False, _is_init=False):
@@ -326,26 +375,38 @@ def get_callable_argspec(fn, no_self=False, _is_init=False):
         raise TypeError("Can't inspect builtin: %s" % fn)
     elif inspect.isfunction(fn):
         if _is_init and no_self:
-            spec = compat.inspect_getargspec(fn)
-            return compat.ArgSpec(
-                spec.args[1:], spec.varargs, spec.keywords, spec.defaults
+            spec = compat.inspect_getfullargspec(fn)
+            return compat.FullArgSpec(
+                spec.args[1:],
+                spec.varargs,
+                spec.varkw,
+                spec.defaults,
+                spec.kwonlyargs,
+                spec.kwonlydefaults,
+                spec.annotations,
             )
         else:
-            return compat.inspect_getargspec(fn)
+            return compat.inspect_getfullargspec(fn)
     elif inspect.ismethod(fn):
         if no_self and (_is_init or fn.__self__):
-            spec = compat.inspect_getargspec(fn.__func__)
-            return compat.ArgSpec(
-                spec.args[1:], spec.varargs, spec.keywords, spec.defaults
+            spec = compat.inspect_getfullargspec(fn.__func__)
+            return compat.FullArgSpec(
+                spec.args[1:],
+                spec.varargs,
+                spec.varkw,
+                spec.defaults,
+                spec.kwonlyargs,
+                spec.kwonlydefaults,
+                spec.annotations,
             )
         else:
-            return compat.inspect_getargspec(fn.__func__)
+            return compat.inspect_getfullargspec(fn.__func__)
     elif inspect.isclass(fn):
         return get_callable_argspec(
             fn.__init__, no_self=no_self, _is_init=True
         )
     elif hasattr(fn, "__func__"):
-        return compat.inspect_getargspec(fn.__func__)
+        return compat.inspect_getfullargspec(fn.__func__)
     elif hasattr(fn, "__call__"):
         if inspect.ismethod(fn.__call__):
             return get_callable_argspec(fn.__call__, no_self=no_self)
@@ -390,8 +451,8 @@ def format_argspec_plus(fn, grouped=True):
     if compat.callable(fn):
         spec = compat.inspect_getfullargspec(fn)
     else:
-        # we accept an existing argspec...
         spec = fn
+
     args = compat.inspect_formatargspec(*spec)
     if spec[0]:
         self_arg = spec[0][0]
@@ -400,22 +461,15 @@ def format_argspec_plus(fn, grouped=True):
     else:
         self_arg = None
 
-    if compat.py3k:
-        apply_pos = compat.inspect_formatargspec(
-            spec[0], spec[1], spec[2], None, spec[4]
-        )
-        num_defaults = 0
-        if spec[3]:
-            num_defaults += len(spec[3])
-        if spec[4]:
-            num_defaults += len(spec[4])
-        name_args = spec[0] + spec[4]
-    else:
-        apply_pos = compat.inspect_formatargspec(spec[0], spec[1], spec[2])
-        num_defaults = 0
-        if spec[3]:
-            num_defaults += len(spec[3])
-        name_args = spec[0]
+    apply_pos = compat.inspect_formatargspec(
+        spec[0], spec[1], spec[2], None, spec[4]
+    )
+    num_defaults = 0
+    if spec[3]:
+        num_defaults += len(spec[3])
+    if spec[4]:
+        num_defaults += len(spec[4])
+    name_args = spec[0] + spec[4]
 
     if num_defaults:
         defaulted_vals = name_args[0 - num_defaults :]
@@ -479,7 +533,7 @@ def getargspec_init(method):
 
     """
     try:
-        return compat.inspect_getargspec(method)
+        return compat.inspect_getfullargspec(method)
     except TypeError:
         if method is object.__init__:
             return (["self"], None, None, None)
@@ -516,30 +570,30 @@ def generic_repr(obj, additional_kw=(), to_inspect=None, omit_kwarg=()):
     vargs = None
     for i, insp in enumerate(to_inspect):
         try:
-            (_args, _vargs, vkw, defaults) = compat.inspect_getargspec(
-                insp.__init__
-            )
+            spec = compat.inspect_getfullargspec(insp.__init__)
         except TypeError:
             continue
         else:
-            default_len = defaults and len(defaults) or 0
+            default_len = spec.defaults and len(spec.defaults) or 0
             if i == 0:
-                if _vargs:
-                    vargs = _vargs
+                if spec.varargs:
+                    vargs = spec.varargs
                 if default_len:
-                    pos_args.extend(_args[1:-default_len])
+                    pos_args.extend(spec.args[1:-default_len])
                 else:
-                    pos_args.extend(_args[1:])
+                    pos_args.extend(spec.args[1:])
             else:
                 kw_args.update(
-                    [(arg, missing) for arg in _args[1:-default_len]]
+                    [(arg, missing) for arg in spec.args[1:-default_len]]
                 )
 
             if default_len:
                 kw_args.update(
                     [
                         (arg, default)
-                        for arg, default in zip(_args[-default_len:], defaults)
+                        for arg, default in zip(
+                            spec.args[-default_len:], spec.defaults
+                        )
                     ]
                 )
     output = []
@@ -654,10 +708,10 @@ def class_hierarchy(cls):
 
 def iterate_attributes(cls):
     """iterate all the keys and attributes associated
-       with a class, without using getattr().
+    with a class, without using getattr().
 
-       Does not use getattr() so that class-sensitive
-       descriptors (i.e. property.__get__()) are not called.
+    Does not use getattr() so that class-sensitive
+    descriptors (i.e. property.__get__()) are not called.
 
     """
     keys = dir(cls)
@@ -710,7 +764,7 @@ def monkeypatch_proxied_specials(
         except AttributeError:
             continue
         try:
-            spec = compat.inspect_getargspec(fn)
+            spec = compat.inspect_getfullargspec(fn)
             fn_args = compat.inspect_formatargspec(spec[0])
             d_args = compat.inspect_formatargspec(spec[0][1:])
         except TypeError:
@@ -948,123 +1002,6 @@ def dependency_for(modulename, add_to_all=False):
     return decorate
 
 
-class dependencies(object):
-    """Apply imported dependencies as arguments to a function.
-
-    E.g.::
-
-        @util.dependencies(
-            "sqlalchemy.sql.widget",
-            "sqlalchemy.engine.default"
-        );
-        def some_func(self, widget, default, arg1, arg2, **kw):
-            # ...
-
-    Rationale is so that the impact of a dependency cycle can be
-    associated directly with the few functions that cause the cycle,
-    and not pollute the module-level namespace.
-
-    """
-
-    def __init__(self, *deps):
-        self.import_deps = []
-        for dep in deps:
-            tokens = dep.split(".")
-            self.import_deps.append(
-                dependencies._importlater(".".join(tokens[0:-1]), tokens[-1])
-            )
-
-    def __call__(self, fn):
-        import_deps = self.import_deps
-        spec = compat.inspect_getfullargspec(fn)
-
-        spec_zero = list(spec[0])
-        hasself = spec_zero[0] in ("self", "cls")
-
-        for i in range(len(import_deps)):
-            spec[0][i + (1 if hasself else 0)] = "import_deps[%r]" % i
-
-        inner_spec = format_argspec_plus(spec, grouped=False)
-
-        for impname in import_deps:
-            del spec_zero[1 if hasself else 0]
-        spec[0][:] = spec_zero
-
-        outer_spec = format_argspec_plus(spec, grouped=False)
-
-        code = "lambda %(args)s: fn(%(apply_kw)s)" % {
-            "args": outer_spec["args"],
-            "apply_kw": inner_spec["apply_kw"],
-        }
-
-        decorated = eval(code, locals())
-        decorated.__defaults__ = getattr(fn, "im_func", fn).__defaults__
-        return update_wrapper(decorated, fn)
-
-    @classmethod
-    def resolve_all(cls, path):
-        for m in list(dependencies._unresolved):
-            if m._full_path.startswith(path):
-                m._resolve()
-
-    _unresolved = set()
-    _by_key = {}
-
-    class _importlater(object):
-        _unresolved = set()
-
-        _by_key = {}
-
-        def __new__(cls, path, addtl):
-            key = path + "." + addtl
-            if key in dependencies._by_key:
-                return dependencies._by_key[key]
-            else:
-                dependencies._by_key[key] = imp = object.__new__(cls)
-                return imp
-
-        def __init__(self, path, addtl):
-            self._il_path = path
-            self._il_addtl = addtl
-            dependencies._unresolved.add(self)
-
-        @property
-        def _full_path(self):
-            return self._il_path + "." + self._il_addtl
-
-        @memoized_property
-        def module(self):
-            if self in dependencies._unresolved:
-                raise ImportError(
-                    "importlater.resolve_all() hasn't "
-                    "been called (this is %s %s)"
-                    % (self._il_path, self._il_addtl)
-                )
-
-            return getattr(self._initial_import, self._il_addtl)
-
-        def _resolve(self):
-            dependencies._unresolved.discard(self)
-            self._initial_import = compat.import_(
-                self._il_path, globals(), locals(), [self._il_addtl]
-            )
-
-        def __getattr__(self, key):
-            if key == "module":
-                raise ImportError(
-                    "Could not resolve module %s" % self._full_path
-                )
-            try:
-                attr = getattr(self.module, key)
-            except AttributeError:
-                raise AttributeError(
-                    "Module %s has no attribute '%s'" % (self._full_path, key)
-                )
-            self.__dict__[key] = attr
-            return attr
-
-
-# from paste.deploy.converters
 def asbool(obj):
     if isinstance(obj, compat.string_types):
         obj = obj.strip().lower()
@@ -1100,17 +1037,24 @@ def asint(value):
     return int(value)
 
 
-def coerce_kw_type(kw, key, type_, flexi_bool=True):
+def coerce_kw_type(kw, key, type_, flexi_bool=True, dest=None):
     r"""If 'key' is present in dict 'kw', coerce its value to type 'type\_' if
     necessary.  If 'flexi_bool' is True, the string '0' is considered false
     when coercing to boolean.
     """
 
-    if key in kw and not isinstance(kw[key], type_) and kw[key] is not None:
+    if dest is None:
+        dest = kw
+
+    if (
+        key in kw
+        and (not isinstance(type_, type) or not isinstance(kw[key], type_))
+        and kw[key] is not None
+    ):
         if type_ is bool and flexi_bool:
-            kw[key] = asbool(kw[key])
+            dest[key] = asbool(kw[key])
         else:
-            kw[key] = type_(kw[key])
+            dest[key] = type_(kw[key])
 
 
 def constructor_copy(obj, cls, *args, **kw):
@@ -1327,6 +1271,41 @@ class symbol(object):
         finally:
             symbol._lock.release()
 
+    @classmethod
+    def parse_user_argument(
+        cls, arg, choices, name, resolve_symbol_names=False
+    ):
+        """Given a user parameter, parse the parameter into a chosen symbol.
+
+        The user argument can be a string name that matches the name of a
+        symbol, or the symbol object itself, or any number of alternate choices
+        such as True/False/ None etc.
+
+        :param arg: the user argument.
+        :param choices: dictionary of symbol object to list of possible
+         entries.
+        :param name: name of the argument.   Used in an :class:`.ArgumentError`
+         that is raised if the parameter doesn't match any available argument.
+        :param resolve_symbol_names: include the name of each symbol as a valid
+         entry.
+
+        """
+        # note using hash lookup is tricky here because symbol's `__hash__`
+        # is its int value which we don't want included in the lookup
+        # explicitly, so we iterate and compare each.
+        for sym, choice in choices.items():
+            if arg is sym:
+                return sym
+            elif resolve_symbol_names and arg == sym.name:
+                return sym
+            elif arg in choice:
+                return sym
+
+        if arg is None:
+            return None
+
+        raise exc.ArgumentError("Invalid value for '%s': %r" % (name, arg))
+
 
 _creation_order = 1
 
@@ -1402,7 +1381,7 @@ def warn(msg):
 
 
 def warn_limited(msg, args):
-    """Issue a warning with a paramterized string, limiting the number
+    """Issue a warning with a parameterized string, limiting the number
     of registrations.
 
     """
@@ -1411,16 +1390,24 @@ def warn_limited(msg, args):
     warnings.warn(msg, exc.SAWarning, stacklevel=2)
 
 
-def only_once(fn):
+def only_once(fn, retry_on_exception):
     """Decorate the given function to be a no-op after it is called exactly
     once."""
 
     once = [fn]
 
     def go(*arg, **kw):
+        # strong reference fn so that it isn't garbage collected,
+        # which interferes with the event system's expectations
+        strong_fn = fn  # noqa
         if once:
             once_fn = once.pop()
-            return once_fn(*arg, **kw)
+            try:
+                return once_fn(*arg, **kw)
+            except:
+                if retry_on_exception:
+                    once.insert(0, once_fn)
+                raise
 
     return go
 
@@ -1474,8 +1461,8 @@ class EnsureKWArgType(type):
                 m = re.match(fn_reg, key)
                 if m:
                     fn = clsdict[key]
-                    spec = compat.inspect_getargspec(fn)
-                    if not spec.keywords:
+                    spec = compat.inspect_getfullargspec(fn)
+                    if not spec.varkw:
                         clsdict[key] = wrapped = cls._wrap_w_kw(fn)
                         setattr(cls, key, wrapped)
         super(EnsureKWArgType, cls).__init__(clsname, bases, clsdict)
@@ -1512,7 +1499,7 @@ def wrap_callable(wrapper, fn):
 
 
 def quoted_token_parser(value):
-    """Parse a dotted identifier with accomodation for quoted names.
+    """Parse a dotted identifier with accommodation for quoted names.
 
     Includes support for SQL-style double quotes as a literal character.
 
@@ -1553,3 +1540,106 @@ def quoted_token_parser(value):
         idx += 1
 
     return ["".join(token) for token in result]
+
+
+def add_parameter_text(params, text):
+    params = _collections.to_list(params)
+
+    def decorate(fn):
+        doc = fn.__doc__ is not None and fn.__doc__ or ""
+        if doc:
+            doc = inject_param_text(doc, {param: text for param in params})
+        fn.__doc__ = doc
+        return fn
+
+    return decorate
+
+
+def _dedent_docstring(text):
+    split_text = text.split("\n", 1)
+    if len(split_text) == 1:
+        return text
+    else:
+        firstline, remaining = split_text
+    if not firstline.startswith(" "):
+        return firstline + "\n" + textwrap.dedent(remaining)
+    else:
+        return textwrap.dedent(text)
+
+
+def inject_docstring_text(doctext, injecttext, pos):
+    doctext = _dedent_docstring(doctext or "")
+    lines = doctext.split("\n")
+    if len(lines) == 1:
+        lines.append("")
+    injectlines = textwrap.dedent(injecttext).split("\n")
+    if injectlines[0]:
+        injectlines.insert(0, "")
+
+    blanks = [num for num, line in enumerate(lines) if not line.strip()]
+    blanks.insert(0, 0)
+
+    inject_pos = blanks[min(pos, len(blanks) - 1)]
+
+    lines = lines[0:inject_pos] + injectlines + lines[inject_pos:]
+    return "\n".join(lines)
+
+
+def inject_param_text(doctext, inject_params):
+    doclines = doctext.splitlines()
+    lines = []
+
+    to_inject = None
+    while doclines:
+        line = doclines.pop(0)
+        if to_inject is None:
+            m = re.match(r"(\s+):param (?:\\\*\*?)?(.+?):", line)
+            if m:
+                param = m.group(2)
+                if param in inject_params:
+                    # default indent to that of :param: plus one
+                    indent = " " * len(m.group(1)) + " "
+
+                    # but if the next line has text, use that line's
+                    # indentntation
+                    if doclines:
+                        m2 = re.match(r"(\s+)\S", doclines[0])
+                        if m2:
+                            indent = " " * len(m2.group(1))
+
+                    to_inject = indent + inject_params[param]
+        elif line.lstrip().startswith(":param "):
+            lines.append("\n")
+            lines.append(to_inject)
+            lines.append("\n")
+            to_inject = None
+        elif not line.rstrip():
+            lines.append(line)
+            lines.append(to_inject)
+            lines.append("\n")
+            to_inject = None
+        elif line.endswith("::"):
+            # TODO: this still wont cover if the code example itself has blank
+            # lines in it, need to detect those via indentation.
+            lines.append(line)
+            lines.append(
+                doclines.pop(0)
+            )  # the blank line following a code example
+            continue
+        lines.append(line)
+
+    return "\n".join(lines)
+
+
+def repr_tuple_names(names):
+    """Trims a list of strings from the middle and return a string of up to
+    four elements. Strings greater than 11 characters will be truncated"""
+    if len(names) == 0:
+        return None
+    flag = len(names) <= 4
+    names = names[0:4] if flag else names[0:3] + names[-1:]
+    res = ["%s.." % name[:11] if len(name) > 11 else name for name in names]
+    if flag:
+        return ", ".join(res)
+    else:
+        return "%s, ..., %s" % (", ".join(res[0:3]), res[-1])
